const commandLineArgs = require('command-line-args');
const getUsage = require('command-line-usage');
const fs = require('fs');
const fsextra = require('fs-extra');
const path = require('path');
const package = require(path.join(process.cwd(), 'package.json'));
const {execSync} = require('child_process');
const packageJson = require('./package.json');

const definitions = [
    { name: 'help', alias: 'h', type: Boolean, description: 'Display this usage guide.' },
    { name: 'dist', type: String, description: `Dist folder path. Default parameter.`, defaultOption: true },
    { name: 'ssr', alias: 's', type: Boolean, description: 'Enable Server Side Rendering.' },
    { name: 'runtime', alias: 'r', type: String, description: 'Specify app.yaml runtime.' },
    { name: 'instance-class', alias: 'i', type: String, description: 'Specify app.yaml instance_class. See https://cloud.google.com/appengine/docs/standard#second-gen-runtimes for details.' },
    { name: 'min-idle-instances', type: Number, description: 'Specify app.yaml min_idle_instances. It keeps a defined number of minimal running instances for version with traffic. Used to reduce latency.' },
    { name: 'version', alias: 'v', type: Boolean, description: 'Print the current version.' },
];

const usage = getUsage([
    {
        header: 'angular-to-appengine',
        content: 'Adapt angular dist folder to be deployed with Google AppEngine.'
    },
    {
        header: 'Options',
        optionList: definitions
    }
]);


const goFile = `
  package main
  func main() {
  
  }
`;

const packageJsonFile = `
{
  "scripts": {
    "start": "node server/main.js"
  }
}
`;

const appFile = (options) => `
service: default
runtime: ${options['runtime'] != null ? options['runtime'] : 'go123'}

inbound_services:
- warmup

instance_class: ${options['instance-class'] != null ? options['instance-class'] : 'F1'}

${options['min-idle-instances'] != null ? `automatic_scaling:\n  min_idle_instances: ${options['min-idle-instances']}`: ''}

handlers:
- url: /(.*\.(appcache|manifest|APPCACHE|MANIFEST))
  mime_type: text/cache-manifest
  static_files: statics/\\1
  upload: statics/(.*\.(appcache|manifest|APPCACHE|MANIFEST))
  secure: always

- url: /(.*\.(atom|ATOM))
  mime_type: application/atom+xml
  static_files: statics/\\1
  upload: statics/(.*\.(atom|ATOM))
  secure: always

- url: /(.*\.(crx|CRX))
  mime_type: application/x-chrome-extension
  static_files: statics/\\1
  upload: statics/(.*\.(crx|CRX))
  secure: always

- url: /(.*\.(css|CSS))
  mime_type: text/css
  static_files: statics/\\1
  upload: statics/(.*\.(css|CSS))
  secure: always

- url: /(.*\.(eot|EOT))
  mime_type: application/vnd.ms-fontobject
  static_files: statics/\\1
  upload: statics/(.*\.(eot|EOT))
  secure: always

- url: /(.*\.(htc|HTC))
  mime_type: text/x-component
  static_files: statics/\\1
  upload: statics/(.*\.(htc|HTC))
  secure: always

- url: /(.*\.(html|HTML))
  mime_type: text/html
  static_files: statics/\\1
  upload: statics/(.*\.(html|HTML))
  secure: always

- url: /(.*\.(ico|ICO))
  mime_type: image/x-icon
  static_files: statics/\\1
  upload: statics/(.*\.(ico|ICO))
  secure: always

- url: /(.*\.(js|JS))
  mime_type: text/javascript
  static_files: statics/\\1
  upload: statics/(.*\.(js|JS))
  secure: always

- url: /(.*.(map|MAP))
  mime_type: application/json
  static_files: statics/\\1
  upload: statics/(.*.(map|MAP))
  secure: always

- url: /(.*\.(json|JSON))
  mime_type: application/json
  static_files: statics/\\1
  upload: statics/(.*\.(json|JSON))
  secure: always
  
# For Cesium only
- url: /(.*\.(czml|CZML))
  mime_type: application/json
  static_files: statics/\\1
  upload: statics/(.*\.(czml|CZML))
  secure: always

- url: /(.*\.(m4v|M4V))
  mime_type: video/m4v
  static_files: statics/\\1
  upload: statics/(.*\.(m4v|M4V))
  secure: always

- url: /(.*\.(mp4|MP4))
  mime_type: video/mp4
  static_files: statics/\\1
  upload: statics/(.*\.(mp4|MP4))
  secure: always

- url: /(.*\.(ogg|oga|OGG|OGA))
  mime_type: audio/ogg
  static_files: statics/\\1
  upload: statics/(.*\.(ogg|oga|OGG|OGA))
  secure: always

- url: /(.*\.(ogv|OGV))
  mime_type: video/ogg
  static_files: statics/\\1
  upload: statics/(.*\.(ogv|OGV))
  secure: always

- url: /(.*\.(otf|OTF))
  mime_type: font/opentype
  static_files: statics/\\1
  upload: statics/(.*\.(otf|OTF))
  secure: always

- url: /(.*\.(rss|RSS))
  mime_type: application/rss+xml
  static_files: statics/\\1
  upload: statics/(.*\.(rss|RSS))
  secure: always

- url: /(.*\.(safariextz|SAFARIEXTZ))
  mime_type: application/octet-stream
  static_files: statics/\\1
  upload: statics/(.*\.(safariextz|SAFARIEXTZ))
  secure: always

- url: /(.*\.(svg|svgz|SVG|SVGZ))
  mime_type: image/svg+xml
  static_files: statics/\\1
  upload: statics/(.*\.(svg|svgz|SVG|SVGZ))
  secure: always

- url: /(.*\.(swf|SWF))
  mime_type: application/x-shockwave-flash
  static_files: statics/\\1
  upload: statics/(.*\.(swf|SWF))
  secure: always

- url: /(.*\.(ttf|TTF))
  mime_type: font/truetype
  static_files: statics/\\1
  upload: statics/(.*\.(ttf|TTF))
  secure: always

- url: /(.*\.(txt|TXT))
  mime_type: text/plain
  static_files: statics/\\1
  upload: statics/(.*\.(txt|TXT))
  secure: always

- url: /(.*.(md|MD))
  mime_type: text/markdown
  static_files: statics/\\1
  upload: statics/(.*.(md|MD))
  secure: always

- url: /(.*\.(unity3d|UNITY3D|unityweb|UNITYWEB))
  mime_type: application/vnd.unity
  static_files: statics/\\1
  upload: statics/(.*\.(unity3d|UNITY3D|unityweb|UNITYWEB))
  secure: always

- url: /(.*\.(webm|WEBM))
  mime_type: video/webm
  static_files: statics/\\1
  upload: statics/(.*\.(webm|WEBM))
  secure: always

- url: /(.*\.(webp|WEBP))
  mime_type: image/webp
  static_files: statics/\\1
  upload: statics/(.*\.(webp|WEBP))
  secure: always

- url: /(.*\.(woff|WOFF))
  mime_type: application/x-font-woff
  static_files: statics/\\1
  upload: statics/(.*\.(woff|WOFF))
  secure: always
  
- url: /(.*\.(woff2|WOFF2))
  mime_type: font/woff2
  static_files: statics/\\1
  upload: statics/(.*\.(woff2|WOFF2))
  secure: always

- url: /(.*\.(xml|XML))
  mime_type: application/xml
  static_files: statics/\\1
  upload: statics/(.*\.(xml|XML))
  secure: always
  
- url: /(.*\.(xsl|XSL))
  mime_type: text/xml
  static_files: statics/\\1
  upload: statics/(.*\.(xsl|XSL))
  secure: always

- url: /(.*\.(xpi|XPI))
  mime_type: application/x-xpinstall
  static_files: statics/\\1
  upload: statics/(.*\.(xpi|XPI))
  secure: always

- url: /(.*.(pdf|PDF))
  mime_type: application/pdf
  static_files: statics/\\1
  upload: statics/(.*.(pdf|PDF))
  secure: always

# image files
- url: /(.*\.(bmp|gif|ico|jpeg|jpg|png|BMP|GIF|ICO|JPEG|JPG|PNG))
  static_files: statics/\\1
  upload: statics/(.*\.(bmp|gif|ico|jpeg|jpg|png|BMP|GIF|ICO|JPEG|JPG|PNG))
  secure: always

# audio files
- url: /(.*\.(mid|midi|mp3|wav|MID|MIDI|MP3|WAV))
  static_files: statics/\\1
  upload: statics/(.*\.(mid|midi|mp3|wav|MID|MIDI|MP3|WAV))
  secure: always

# windows files
- url: /(.*\.(doc|exe|ppt|rtf|xls|DOC|EXE|PPT|RTF|XLS))
  static_files: statics/\\1
  upload: statics/(.*\.(doc|exe|ppt|rtf|xls|DOC|EXE|PPT|RTF|XLS))
  secure: always

# compressed files
- url: /(.*\.(bz2|gz|rar|tar|tgz|zip|BZ2|GZ|RAR|TAR|TGZ|ZIP))
  static_files: statics/\\1
  upload: statics/(.*\.(bz2|gz|rar|tar|tgz|zip|BZ2|GZ|RAR|TAR|TGZ|ZIP))
  secure: always

- url: /
  static_files: statics/index.html
  upload: statics/index.html
  expiration: "1s"
  secure: always

- url: /info
  static_files: statics/app-info.json
  upload: statics/app-info.json
  expiration: "1s"
  secure: always
  
- url: /info/bundle-stats
  static_files: statics/bundle-stats.html
  upload: statics/bundle-stats.html
  secure: always

- url: /(.*)
  static_files: statics/index.html
  upload: statics/index.html
  expiration: "1s"
  secure: always
`;

const appFileSSR = (options) => `
service: default
runtime: ${options['runtime'] != null ? options['runtime'] : 'nodejs10'}

inbound_services:
- warmup

instance_class: ${options['instance-class'] != null ? options['instance-class'] : 'F1'}

${options['min-idle-instances'] != null ? `automatic_scaling:\n  min_idle_instances: ${options['min-idle-instances']}`: ''}

handlers:
- url: /CHANGELOG.md
  static_files: browser/CHANGELOG.md
  upload: browser/CHANGELOG.md
  expiration: "1s"
  secure: always
  
- url: /info
  static_files: browser/app-info.json
  upload: browser/app-info.json
  expiration: "1s"
  secure: always
  
- url: /.*
  script: auto
  secure: always
  redirect_http_response_code: 301
`;

const lastCommitInfo = () => {
    const run = command => execSync(command, {encoding: 'utf8'});
    const prettyFormat = ["%h", "%H", "%s", "%f", "%b", "%at", "%ct", "%an", "%ae", "%cn", "%ce", "%N"];
    const logCmd = `git log -1 --pretty=format:"${prettyFormat.join(',')}" && git rev-parse --abbrev-ref HEAD`;
    const statusCmd = 'git status -s';

    const [
        shortHash,
        hash,
        subject,
        sanitizedSubject,
        body,
        authoredOn,
        committedOn,
        authorName,
        authorEmail,
        committerName,
        committerEmail,
        notes,
        branch
    ] = run(logCmd).trim().split(',');

    const isDirty = run(statusCmd).trim().length !== 0;

    return {
        shortHash: shortHash,
        hash: hash,
        subject: subject,
        sanitizedSubject: sanitizedSubject,
        body: body,
        authoredOn: parseInt(authoredOn),
        committedOn: parseInt(committedOn),
        author: {
            name: authorName,
            email: authorEmail,
        },
        committer: {
            name: committerName,
            email: committerEmail
        },
        notes: notes,
        branch: branch === undefined ? hash : branch,
        dirty: isDirty
    };
};

const appInfo = () => {
    const lastCommit = lastCommitInfo();

    return {
        app: {
            name: package.name,
            version: package.version
        },
        git: {
            commit: {
                time: lastCommit.committedOn,
                id: lastCommit.shortHash,
                hash: lastCommit.hash,
                dirty: lastCommit.dirty
            },
            branch: lastCommit.branch
        }
    };
};

exports.angularToAppengine = function () {

    const options = commandLineArgs(definitions, { stopAtFirstUnknown: true });

    if (options['version']) {
        console.log(packageJson.version);
        process.exit();
    }

    if (options['help']) {
        console.log(usage);
        process.exit();
    }

    let distSubFolder = "./dist";
    if (options['dist'] != null) {
        distSubFolder = options['dist'];
    }

    let dist = path.join(process.cwd(), distSubFolder);
    const tmpStatics = path.join(process.cwd(), '/statics');
    const finalStatics = path.join(dist, '/statics');
    const changelogFilename = 'CHANGELOG.md';

    if (!options["ssr"]) {
        if (fs.existsSync(path.join(process.cwd(), changelogFilename))) {
            fs.copyFileSync(path.join(process.cwd(), changelogFilename), path.join(dist, changelogFilename));
        }

        fs.writeFileSync(path.join(dist, 'app-info.json'), JSON.stringify(appInfo()));

        fsextra.mkdirpSync(tmpStatics);
        fsextra.moveSync(dist, tmpStatics, {overwrite: true});

        fsextra.mkdirpSync(dist);
        fs.writeFileSync(path.join(dist, 'main.go'), goFile);
        fs.writeFileSync(path.join(dist, 'app.yaml'), appFile(options));

        fsextra.moveSync(tmpStatics, finalStatics, {overwrite: true});
    } else {
        if (fs.existsSync(path.join(process.cwd(), changelogFilename))) {
            fs.copyFileSync(path.join(process.cwd(), changelogFilename), path.join(dist, 'browser', changelogFilename));
        }

        fs.writeFileSync(path.join(dist, 'browser', 'app-info.json'), JSON.stringify(appInfo()));
        fs.writeFileSync(path.join(dist, 'package.json'), packageJsonFile);
        fs.writeFileSync(path.join(dist, 'app.yaml'), appFileSSR(options));
    }
}
